/*
 * Copyright (c) 2006, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * Licensed under the Universal Permissive License v 1.0 as shown at
 * http://oss.oracle.com/licenses/upl.
 *
 * Licensed under the GNU General Public License (GPL), version 2. See the file
 * COPYING in the top level of this tree.
 */

#include <string.h>
#include <sys/types.h>
#include <assert.h>
#include <gelf.h>
#include <ctf_impl.h>
#include <sys/mman.h>
#include <zlib.h>

static const ctf_dmodel_t _libctf_models[] = {
	{ "ILP32", CTF_MODEL_ILP32, 4, 1, 2, 4, 4 },
	{ "LP64", CTF_MODEL_LP64, 8, 1, 2, 4, 8 },
	{ NULL, 0, 0, 0, 0, 0, 0 }
};

const char _CTF_SECTION[] = ".ctf";
const char _CTF_NULLSTR[] = "";

int _libctf_version = CTF_VERSION;	/* library client version */
int _libctf_debug = 0;			/* debugging messages enabled */

static uint32_t
get_kind_v1(uint32_t info)
{
	return (CTF_V1_INFO_KIND(info));
}

static uint32_t
get_root_v1(uint32_t info)
{
	return (CTF_V1_INFO_ISROOT(info));
}

static uint32_t
get_vlen_v1(uint32_t info)
{
	return (CTF_V1_INFO_VLEN(info));
}

static uint32_t
get_kind_v2(uint32_t info)
{
	return (CTF_V2_INFO_KIND(info));
}

static uint32_t
get_root_v2(uint32_t info)
{
	return (CTF_V2_INFO_ISROOT(info));
}

static uint32_t
get_vlen_v2(uint32_t info)
{
	return (CTF_V2_INFO_VLEN(info));
}

static inline ssize_t
get_ctt_size_common(const ctf_file_t *fp, const ctf_type_t *tp,
    ssize_t *sizep, ssize_t *incrementp, size_t lsize, size_t csize,
    size_t ctf_type_size, size_t ctf_stype_size,
    size_t ctf_lsize_sent)
{
	ssize_t size, increment;

	if (csize == ctf_lsize_sent) {
		size = lsize;
		increment = ctf_type_size;
	} else {
		size = csize;
		increment = ctf_stype_size;
	}

	if (sizep)
		*sizep = size;
	if (incrementp)
		*incrementp = increment;

	return (size);
}

static ssize_t
get_ctt_size_v1(const ctf_file_t *fp, const ctf_type_t *tp,
    ssize_t *sizep, ssize_t *incrementp)
{
	ctf_type_v1_t *t1p = (ctf_type_v1_t *) tp;

	return get_ctt_size_common(fp, tp, sizep, incrementp,
	    CTF_TYPE_LSIZE(t1p), t1p->ctt_size, sizeof (ctf_type_v1_t),
	    sizeof (ctf_stype_v1_t), CTF_LSIZE_SENT_V1);
}

/*
 * Return the size that a v1 will be once it is converted to v2.
 */
static ssize_t
get_ctt_size_v2_unconverted(const ctf_file_t *fp, const ctf_type_t *tp,
    ssize_t *sizep, ssize_t *incrementp)
{
	ctf_type_v1_t *t1p = (ctf_type_v1_t *) tp;

	return get_ctt_size_common(fp, tp, sizep, incrementp,
	    CTF_TYPE_LSIZE(t1p), t1p->ctt_size, sizeof (ctf_type_t),
	    sizeof (ctf_stype_t), CTF_LSIZE_SENT);
}

static ssize_t
get_ctt_size_v2(const ctf_file_t *fp, const ctf_type_t *tp,
    ssize_t *sizep, ssize_t *incrementp)
{
	return get_ctt_size_common(fp, tp, sizep, incrementp,
	    CTF_TYPE_LSIZE(tp), tp->ctt_size, sizeof (ctf_type_t),
	    sizeof (ctf_stype_t), CTF_LSIZE_SENT);
}

static ssize_t get_vbytes_common(ushort_t kind, ssize_t size, size_t vlen)
{
	switch (kind) {
	case CTF_K_INTEGER: case CTF_K_FLOAT:
		return (sizeof (uint32_t));
	case CTF_K_ENUM:
		return (sizeof (ctf_enum_t) * vlen);
	case CTF_K_FORWARD: case CTF_K_UNKNOWN: case CTF_K_POINTER:
	case CTF_K_TYPEDEF: case CTF_K_VOLATILE: case CTF_K_CONST:
	case CTF_K_RESTRICT:
		return (0);
	default:
		ctf_dprintf("detected invalid CTF kind -- %u\n", kind);
		return (ECTF_CORRUPT);
	}
}

static ssize_t
get_vbytes_v1(ushort_t kind, ssize_t size, size_t vlen)
{
	switch (kind) {
	case CTF_K_ARRAY:
		return (sizeof (ctf_array_v1_t));
	case CTF_K_FUNCTION:
		return (sizeof (ushort_t) * (vlen + (vlen & 1)));
	case CTF_K_STRUCT: case CTF_K_UNION:
		if (size < CTF_LSTRUCT_THRESH_V1)
			return (sizeof (ctf_member_v1_t) * vlen);
		else
			return (sizeof (ctf_lmember_v1_t) * vlen);
	}

	return get_vbytes_common(kind, size, vlen);
}

static ssize_t
get_vbytes_v2(ushort_t kind, ssize_t size, size_t vlen)
{
	switch (kind) {
	case CTF_K_ARRAY:
		return (sizeof (ctf_array_t));
	case CTF_K_FUNCTION:
		return (sizeof (uint32_t) * (vlen + (vlen & 1)));
	case CTF_K_STRUCT: case CTF_K_UNION:
		if (size < CTF_LSTRUCT_THRESH)
			return (sizeof (ctf_member_t) * vlen);
		else
			return (sizeof (ctf_lmember_t) * vlen);
	}

	return get_vbytes_common(kind, size, vlen);
}

static const ctf_fileops_t ctf_fileops[] = {
	{ NULL, NULL, NULL },
	/* CTF_VERSION_1 */
	{ get_kind_v1, get_root_v1, get_vlen_v1, get_ctt_size_v1, get_vbytes_v1 },
	/* CTF_VERSION_1_UPGRADED_2 */
	{ get_kind_v2, get_root_v2, get_vlen_v2, get_ctt_size_v2, get_vbytes_v2 },
	/* CTF_VERSION_2 */
	{ get_kind_v2, get_root_v2, get_vlen_v2, get_ctt_size_v2, get_vbytes_v2 },
};

/*
 * Convert a 32-bit ELF symbol into GElf (Elf64) and return a pointer to it.
 */
static Elf64_Sym *
sym_to_gelf(const Elf32_Sym *src, Elf64_Sym *dst)
{
	dst->st_name = src->st_name;
	dst->st_value = src->st_value;
	dst->st_size = src->st_size;
	dst->st_info = src->st_info;
	dst->st_other = src->st_other;
	dst->st_shndx = src->st_shndx;

	return (dst);
}

/*
 * Initialize the symtab translation table by filling each entry with the
 * offset of the CTF type or function data corresponding to each STT_FUNC or
 * STT_OBJECT entry in the symbol table.
 */
static int
init_symtab(ctf_file_t *fp, const ctf_header_t *hp,
    const ctf_sect_t *sp, const ctf_sect_t *strp)
{
	const uchar_t *symp = sp->cts_data;
	uint32_t *xp = fp->ctf_sxlate;
	uint32_t *xend = xp + fp->ctf_nsyms;

	uint32_t objtoff = hp->cth_objtoff;
	uint32_t funcoff = hp->cth_funcoff;

	ushort_t info, vlen;
	Elf64_Sym sym, *gsp;
	const char *name;

	/*
	 * The CTF data object and function type sections are ordered to match
	 * the relative order of the respective symbol types in the symtab.
	 * If no type information is available for a symbol table entry, a
	 * pad is inserted in the CTF section.  As a further optimization,
	 * anonymous or undefined symbols are omitted from the CTF data.
	 */
	for (; xp < xend; xp++, symp += sp->cts_entsize) {
		if (sp->cts_entsize == sizeof (Elf32_Sym))
			gsp = sym_to_gelf((Elf32_Sym *)(uintptr_t)symp, &sym);
		else
			gsp = (Elf64_Sym *)(uintptr_t)symp;

		if (gsp->st_name < strp->cts_size)
			name = (const char *)strp->cts_data + gsp->st_name;
		else
			name = _CTF_NULLSTR;

		if (gsp->st_name == 0 || gsp->st_shndx == SHN_UNDEF ||
		    strcmp(name, "_START_") == 0 ||
		    strcmp(name, "_END_") == 0) {
			*xp = -1u;
			continue;
		}

		switch (ELF64_ST_TYPE(gsp->st_info)) {
		case STT_OBJECT:
			if (objtoff >= hp->cth_funcoff ||
			    (gsp->st_shndx == SHN_ABS && gsp->st_value == 0)) {
				*xp = -1u;
				break;
			}

			*xp = objtoff;
			objtoff += sizeof (uint32_t);
			break;

		case STT_FUNC:
			if (funcoff >= hp->cth_typeoff) {
				*xp = -1u;
				break;
			}

			*xp = funcoff;

			info = *(uint32_t *)((uintptr_t)fp->ctf_buf + funcoff);
			vlen = LCTF_INFO_VLEN(fp, info);

			/*
			 * If we encounter a zero pad at the end, just skip it.
			 * Otherwise skip over the function and its return type
			 * (+2) and the argument list (vlen).
			 */
			if (LCTF_INFO_KIND(fp, info) == CTF_K_UNKNOWN &&
			    vlen == 0)
				funcoff += sizeof (uint32_t); /* skip pad */
			else
				funcoff += sizeof (uint32_t) * (vlen + 2);
			break;

		default:
			*xp = -1u;
			break;
		}
	}

	ctf_dprintf("loaded %lu symtab entries\n", fp->ctf_nsyms);
	return (0);
}

/*
 * Set the CTF base pointer and derive the buf pointer from it, initializing
 * everything in the ctf_file that depends on the base or buf pointers.
 */
static void
ctf_set_base(ctf_file_t *fp, const ctf_header_t *hp, void *base)
{
	fp->ctf_base = base;
	fp->ctf_buf = fp->ctf_base + sizeof (ctf_header_t);
	fp->ctf_vars = (ctf_varent_t *)((const char *)fp->ctf_buf + hp->cth_varoff);
	fp->ctf_nvars = (hp->cth_typeoff - hp->cth_varoff) / sizeof (ctf_varent_t);

	fp->ctf_str[CTF_STRTAB_0].cts_strs = (const char *) fp->ctf_buf +
	    hp->cth_stroff;
	fp->ctf_str[CTF_STRTAB_0].cts_len = hp->cth_strlen;

	/*
	 * If we have a parent container name and label, store the relocated
	 * string pointers in the CTF container for easy access later.
	 *
	 * Note: before conversion, these will be set to values that will be
	 * immediately invalidated by the conversion process, but the conversion
	 * process will call ctf_set_base() again to fix things up.
	 */
	if (hp->cth_parlabel != 0)
		fp->ctf_parlabel = ctf_strptr(fp, hp->cth_parlabel);
	if (hp->cth_parname != 0)
		fp->ctf_parname = ctf_strptr(fp, hp->cth_parname);

	ctf_dprintf("ctf_set_base: parent name %s (label %s)\n",
	    fp->ctf_parname ? fp->ctf_parname : "<NULL>",
	    fp->ctf_parlabel ? fp->ctf_parlabel : "<NULL>");
}

/*
 * Free a ctf_base pointer: the pointer passed, or (if NULL) fp->ctf_base.
 */
static void
ctf_free_base(ctf_file_t *fp, uchar_t *ctf_base, size_t ctf_size)
{
	uchar_t *base;
	size_t size;

	if (ctf_base) {
		base = ctf_base;
		size = ctf_size;
	} else {
		base = (uchar_t *)fp->ctf_base;
		size = fp->ctf_size;
	}

	if (base != fp->ctf_data.cts_data && base != NULL)
		ctf_data_free(base, size);
}

/*
 * Set the version of the CTF file.
 *
 * When this is reset, LCTF_* changes behaviour, but there is no guarantee that
 * the variable data list associated with each type has been upgraded: this is
 * tracked on a per-type basis via bit vectors such as ctf_v1 in the fp.
 */
static void
ctf_set_version(ctf_file_t *fp, ctf_header_t *cth, int ctf_version)
{
	fp->ctf_version = ctf_version;
	cth->cth_version = ctf_version;
	fp->ctf_fileops = &ctf_fileops[ctf_version];
}

/*
 * Upgrade the type table to CTF_VERSION_2 (really CTF_VERSION_1_UPGRADED_2).
 *
 * The upgrade is not done in-place: the ctf_base is moved.  ctf_strptr() must
 * not be called before reallocation is complete.
 */
static int
upgrade_types(ctf_file_t *fp, ctf_header_t *cth)
{
	const ctf_type_v1_t *tbuf;
	const ctf_type_v1_t *tend;
	uchar_t *ctf_base, *old_ctf_base = (uchar_t *)fp->ctf_base;
	size_t old_ctf_size = fp->ctf_size;
	ctf_type_t *t2buf;

	ssize_t increase = 0, size, increment, v2increment, vbytes, v2bytes;
	const ctf_type_v1_t *tp;
	ctf_type_t *t2p;
	ctf_header_t *new_cth;

	tbuf = (ctf_type_v1_t *)(fp->ctf_buf + cth->cth_typeoff);
	tend = (ctf_type_v1_t *)(fp->ctf_buf + cth->cth_stroff);

	/*
	 * Much like init_types(), this is a two-pass process.
	 *
	 * First, figure out the new type-section size needed.  (It is possible,
	 * in theory, for it to be less than the old size, but this is very
	 * unlikely.  It cannot be so small that cth_typeoff ends up of negative
	 * size.  We validate this with an assertion below.)
	 *
	 * We must cater not only for changes in vlen and types sizes but also
	 * for changes in 'increment', which happen because v2 places some types
	 * into ctf_stype_t where v1 would be forced to use the larger non-stype.
	 */
	for (tp = tbuf; tp < tend;
	     tp = (ctf_type_v1_t *)((uintptr_t)tp + increment + vbytes)) {
		ushort_t kind = CTF_V1_INFO_KIND(tp->ctt_info);
		ulong_t vlen = CTF_V1_INFO_VLEN(tp->ctt_info);

		size = get_ctt_size_v1(fp, (const ctf_type_t *)tp, NULL,
		    &increment);
		vbytes = get_vbytes_v1(kind, size, vlen);

		get_ctt_size_v2_unconverted(fp, (const ctf_type_t *)tp, NULL,
		    &v2increment);
		v2bytes = get_vbytes_v2(kind, size, vlen);

		if ((vbytes < 0) || (size < 0))
			return (ECTF_CORRUPT);

		increase += v2increment - increment; /* may be negative */
		increase += v2bytes - vbytes;
	}

	/*
	 * Allocate enough room for the new buffer, then copy everything but the
	 * type section into place, and reset the base accordingly.  Leave the
	 * version number unchanged, so that LCTF_INFO_* still works on the
	 * as-yet-untranslated type info.
	 */
	if ((ctf_base = ctf_data_alloc(fp->ctf_size + increase)) == MAP_FAILED)
		return (ECTF_ZALLOC);

	bcopy(fp->ctf_base, ctf_base, sizeof (ctf_header_t) + cth->cth_typeoff);
	bcopy(fp->ctf_base + sizeof (ctf_header_t) + cth->cth_stroff,
	    ctf_base + sizeof (ctf_header_t) + cth->cth_stroff + increase,
	    cth->cth_strlen);

	bzero(ctf_base + sizeof (ctf_header_t) + cth->cth_typeoff,
	    cth->cth_stroff - cth->cth_typeoff + increase);

	/*
	 * The cth here is an automatic variable in ctf_bufopen(), and transient
	 * (a copy maintained because at that stage the header read out of the
	 * ctf file may be read-only). We make all modifications in the
	 * canonical copy at ctf_base (by now, writable), then copy it back into
	 * cth at the end.
	 */

	new_cth = (ctf_header_t *) ctf_base;
	new_cth->cth_stroff += increase;
	fp->ctf_size += increase;
	assert(new_cth->cth_stroff >= new_cth->cth_typeoff);
	ctf_set_base(fp, new_cth, ctf_base);

	t2buf = (ctf_type_t *)(fp->ctf_buf + new_cth->cth_typeoff);

	/*
	 * Iterate through all the types again, upgrading them.
	 *
	 * Everything that hasn't changed can just be outright bcopy()ed.
	 * Things that have changed need field-by-field consideration.
	 */
	for (tp = tbuf, t2p = t2buf; tp < tend;
	     tp = (ctf_type_v1_t *)((uintptr_t)tp + increment + vbytes),
		 t2p = (ctf_type_t *)((uintptr_t)t2p + v2increment + v2bytes)) {
		ushort_t kind = CTF_V1_INFO_KIND(tp->ctt_info);
		int isroot = CTF_V1_INFO_ISROOT(tp->ctt_info);
		ulong_t vlen = CTF_V1_INFO_VLEN(tp->ctt_info);
		ssize_t v2size;
		void *vdata, *v2data;

		size = get_ctt_size_v1(fp, (const ctf_type_t *)tp, NULL,
		    &increment);
		vbytes = get_vbytes_v1(kind, size, vlen);

		t2p->ctt_name = tp->ctt_name;
		t2p->ctt_info = CTF_TYPE_INFO(kind, isroot, vlen);

		switch (kind) {
		case CTF_K_FUNCTION:
		case CTF_K_FORWARD:
		case CTF_K_TYPEDEF:
		case CTF_K_POINTER:
		case CTF_K_VOLATILE:
		case CTF_K_CONST:
		case CTF_K_RESTRICT:
			t2p->ctt_type = tp->ctt_type;
			break;
		case CTF_K_INTEGER:
		case CTF_K_FLOAT:
		case CTF_K_ARRAY:
		case CTF_K_STRUCT:
		case CTF_K_UNION:
		case CTF_K_ENUM:
		case CTF_K_UNKNOWN:
			if (size <= CTF_MAX_SIZE)
				t2p->ctt_size = size;
			else {
				t2p->ctt_lsizehi = CTF_SIZE_TO_LSIZE_HI(size);
				t2p->ctt_lsizelo = CTF_SIZE_TO_LSIZE_LO(size);
			}
		break;
		}

		v2size = get_ctt_size_v2(fp, t2p, NULL, &v2increment);
		v2bytes = get_vbytes_v2(kind, v2size, vlen);

		/*
		 * Catch out-of-sync get_ctt_size_*().  The count goes wrong if
		 * these are not identical (and having them different makes no
		 * sense semantically).
		 */
		assert(size == v2size);

		/*
		 * Now the varlen info.
		 */

		vdata = (void *)((uintptr_t)tp + increment);
		v2data = (void *)((uintptr_t)t2p + v2increment);

		switch (kind) {
		case CTF_K_ARRAY: {
			const ctf_array_v1_t *ap = (const ctf_array_v1_t *)
			    vdata;
			ctf_array_t *a2p = (ctf_array_t *) v2data;

			a2p->cta_contents = ap->cta_contents;
			a2p->cta_index = ap->cta_index;
			a2p->cta_nelems = ap->cta_nelems;
			break;
		}
		case CTF_K_STRUCT:
		case CTF_K_UNION: {
			ctf_member_t tmp;
			const ctf_member_v1_t *m1 = (const ctf_member_v1_t *)
			    vdata;
			const ctf_lmember_v1_t *lm1 = (const ctf_lmember_v1_t *)
			    m1;
			ctf_member_t *m2 = (ctf_member_t *) v2data;
			ctf_lmember_t *lm2 = (ctf_lmember_t *) m2;
			ulong_t i;

			/*
			 * We walk all four pointers forward, but only reference
			 * the two that are valid for the given size, to avoid
			 * quadruplicating all the code.
			 */

			for (i = vlen; i != 0; i--, m1++, lm1++, m2++, lm2++) {
				size_t offset;
				if (size < CTF_LSTRUCT_THRESH_V1) {
					offset = m1->ctm_offset;
					tmp.ctm_name = m1->ctm_name;
					tmp.ctm_type = m1->ctm_type;
				} else {
					offset = CTF_LMEM_OFFSET(lm1);
					tmp.ctm_name = lm1->ctlm_name;
					tmp.ctm_type = lm1->ctlm_type;
				}
				if (size < CTF_LSTRUCT_THRESH) {
					m2->ctm_name = tmp.ctm_name;
 					m2->ctm_type = tmp.ctm_type;
					m2->ctm_offset = offset;
				} else {
					lm2->ctlm_name = tmp.ctm_name;
 					lm2->ctlm_type = tmp.ctm_type;
					lm2->ctlm_offsethi = CTF_OFFSET_TO_LMEMHI(offset);
					lm2->ctlm_offsetlo = CTF_OFFSET_TO_LMEMLO(offset);
				}
			}
			break;
		}
		case CTF_K_FUNCTION: {
			ulong_t i;
			ushort_t *a1 = (ushort_t *) vdata;
			uint32_t *a2 = (uint32_t *) v2data;
			for (i = vlen; i != 0; i--, a1++, a2++)
				*a2 = *a1;
		}
		default:
			/*
			 * Catch out-of-sync get_vbytes_*()
			 */
			assert(vbytes == v2bytes);
			bcopy(vdata, v2data, vbytes);
		}
	}
	/*
	 * Verify that the entire region was converted.  If not,
	 * we are either converting too much, or too little (leading to a buffer
	 * overrun either here or at read time, in init_types().)
	 */
	assert((size_t) t2p - (size_t) fp->ctf_buf == new_cth->cth_stroff);

	ctf_set_version(fp, (ctf_header_t *) ctf_base,
	    CTF_VERSION_1_UPGRADED_2);
	ctf_free_base(fp, old_ctf_base, old_ctf_size);
	bcopy(new_cth, cth, sizeof (ctf_header_t));

	return (0);
}

/*
 * Initialize the type ID translation table with the byte offset of each type,
 * and initialize the hash tables of each named type.  Upgrade the type table
 * to the latest supported representation in the process, if needed.
 */
static int
init_types(ctf_file_t *fp, ctf_header_t *cth)
{
	const ctf_type_t *tbuf;
	const ctf_type_t *tend;

	ulong_t pop[CTF_K_MAX + 1] = { 0 };
	const ctf_type_t *tp;
	ctf_hash_t *hp;
	uint32_t id, dst;
	uint32_t *xp;

	/*
	 * We determine whether the container is a child or a parent based on
	 * the value of cth_parname.
	 */
	int child = cth->cth_parname != 0;
	int nlstructs = 0, nlunions = 0;
	int err;

	if (_dt_unlikely_(fp->ctf_version == CTF_VERSION_1)) {
		int err;
		if ((err = upgrade_types(fp, cth)) != 0)
			return (err); 	/* upgrade failed */
	}

	tbuf = (ctf_type_t *)(fp->ctf_buf + cth->cth_typeoff);
	tend = (ctf_type_t *)(fp->ctf_buf + cth->cth_stroff);

	/*
	 * We make two passes through the entire type section.  In this first
	 * pass, we count the number of each type and the total number of types.
	 */
	for (tp = tbuf; tp < tend; fp->ctf_typemax++) {
		ushort_t kind = LCTF_INFO_KIND(fp, tp->ctt_info);
		ulong_t vlen = LCTF_INFO_VLEN(fp, tp->ctt_info);
		ssize_t size, increment, vbytes;

		(void) ctf_get_ctt_size(fp, tp, &size, &increment);
		vbytes = LCTF_VBYTES(fp, kind, size, vlen);

		if (vbytes < 0)
			return (ECTF_CORRUPT);

		if (kind == CTF_K_FORWARD) {
			/*
			 * For forward declarations, ctt_type is the CTF_K_*
			 * kind for the tag, so bump that population count too.
			 * If ctt_type is unknown, treat the tag as a struct.
			 */
			if (tp->ctt_type == CTF_K_UNKNOWN ||
			    tp->ctt_type >= CTF_K_MAX)
				pop[CTF_K_STRUCT]++;
			else
				pop[tp->ctt_type]++;
		}
		tp = (ctf_type_t *)((uintptr_t)tp + increment + vbytes);
		pop[kind]++;
	}

	if (child) {
		ctf_dprintf("CTF container %p is a child\n", (void *)fp);
		fp->ctf_flags |= LCTF_CHILD;
	} else
		ctf_dprintf("CTF container %p is a parent\n", (void *)fp);

	/*
	 * Now that we've counted up the number of each type, we can allocate
	 * the hash tables, type translation table, and pointer table.
	 */
	if ((err = ctf_hash_create(&fp->ctf_structs, pop[CTF_K_STRUCT])) != 0)
		return (err);

	if ((err = ctf_hash_create(&fp->ctf_unions, pop[CTF_K_UNION])) != 0)
		return (err);

	if ((err = ctf_hash_create(&fp->ctf_enums, pop[CTF_K_ENUM])) != 0)
		return (err);

	if ((err = ctf_hash_create(&fp->ctf_names,
	    pop[CTF_K_INTEGER] + pop[CTF_K_FLOAT] + pop[CTF_K_FUNCTION] +
	    pop[CTF_K_TYPEDEF] + pop[CTF_K_POINTER] + pop[CTF_K_VOLATILE] +
	    pop[CTF_K_CONST] + pop[CTF_K_RESTRICT])) != 0)
		return (err);

	fp->ctf_txlate = ctf_alloc(sizeof (uint32_t) * (fp->ctf_typemax + 1));
	fp->ctf_ptrtab = ctf_alloc(sizeof (uint32_t) * (fp->ctf_typemax + 1));

	if (fp->ctf_txlate == NULL || fp->ctf_ptrtab == NULL)
		return (ENOMEM); /* memory allocation failed */

	xp = fp->ctf_txlate;
	*xp++ = 0; /* type id 0 is used as a sentinel value */

	bzero(fp->ctf_txlate, sizeof (uint32_t) * (fp->ctf_typemax + 1));
	bzero(fp->ctf_ptrtab, sizeof (uint32_t) * (fp->ctf_typemax + 1));

	/*
	 * In the second pass through the types, we fill in each entry of the
	 * type and pointer tables and add names to the appropriate hashes.
	 */
	for (id = 1, tp = tbuf; tp < tend; xp++, id++) {
		ushort_t kind = LCTF_INFO_KIND(fp, tp->ctt_info);
		ushort_t flag = LCTF_INFO_ISROOT(fp, tp->ctt_info);
		ulong_t vlen = LCTF_INFO_VLEN(fp, tp->ctt_info);
		ssize_t size, increment, vbytes;

		const char *name;
		ctf_helem_t *hep;

		(void) ctf_get_ctt_size(fp, tp, &size, &increment);
		name = ctf_strptr(fp, tp->ctt_name);
		vbytes = LCTF_VBYTES(fp, kind, size, vlen);

		switch (kind) {
		case CTF_K_INTEGER:
		case CTF_K_FLOAT:
			/*
			 * Names are reused by bit-fields, which are
			 * differentiated by their encodings, and so typically
			 * we'd record only the first instance of a given
			 * intrinsic.  However, we replace an existing type with
			 * a root-visible version so that we can be sure to find
			 * it when checking for conflicting definitions in
			 * ctf_add_type().
			 */
			if ((hep = ctf_hash_lookup(&fp->ctf_names, fp,
			    name, strlen(name))) == NULL) {
				err = ctf_hash_insert(&fp->ctf_names, fp,
				    LCTF_INDEX_TO_TYPE(fp, id, child),
				    tp->ctt_name);
				if (err != 0 && err != ECTF_STRTAB)
					return (err);
			} else if (flag & CTF_ADD_ROOT) {
				hep->h_type = LCTF_INDEX_TO_TYPE(fp, id, child);
			}
			break;

		case CTF_K_ARRAY:
			break;

		case CTF_K_FUNCTION:
			err = ctf_hash_insert(&fp->ctf_names, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);
			if (err != 0 && err != ECTF_STRTAB)
				return (err);
			break;

		case CTF_K_STRUCT:
			err = ctf_hash_define(&fp->ctf_structs, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);

			if (err != 0 && err != ECTF_STRTAB)
				return (err);

			if (size >= CTF_LSTRUCT_THRESH)
				nlstructs++;
			break;

		case CTF_K_UNION:
			err = ctf_hash_define(&fp->ctf_unions, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);

			if (err != 0 && err != ECTF_STRTAB)
				return (err);

			if (size >= CTF_LSTRUCT_THRESH)
				nlunions++;
			break;

		case CTF_K_ENUM:
			err = ctf_hash_define(&fp->ctf_enums, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);

			if (err != 0 && err != ECTF_STRTAB)
				return (err);
			break;

		case CTF_K_TYPEDEF:
			err = ctf_hash_insert(&fp->ctf_names, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);
			if (err != 0 && err != ECTF_STRTAB)
				return (err);
			break;

		case CTF_K_FORWARD:
			/*
			 * Only insert forward tags into the given hash if the
			 * type or tag name is not already present.
			 */
			switch (tp->ctt_type) {
			case CTF_K_STRUCT:
				hp = &fp->ctf_structs;
				break;
			case CTF_K_UNION:
				hp = &fp->ctf_unions;
				break;
			case CTF_K_ENUM:
				hp = &fp->ctf_enums;
				break;
			default:
				hp = &fp->ctf_structs;
			}

			if (ctf_hash_lookup(hp, fp,
			    name, strlen(name)) == NULL) {
				err = ctf_hash_insert(hp, fp,
				    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);
				if (err != 0 && err != ECTF_STRTAB)
					return (err);
			}
			break;

		case CTF_K_POINTER:
			/*
			 * If the type referenced by the pointer is in this CTF
			 * container, then store the index of the pointer type
			 * in fp->ctf_ptrtab[ index of referenced type ].
			 */
			if (LCTF_TYPE_ISCHILD(fp, tp->ctt_type) == child &&
			    LCTF_TYPE_TO_INDEX(fp, tp->ctt_type) <= fp->ctf_typemax)
				fp->ctf_ptrtab[
				    LCTF_TYPE_TO_INDEX(fp, tp->ctt_type)] = id;
			/*FALLTHRU*/

		case CTF_K_VOLATILE:
		case CTF_K_CONST:
		case CTF_K_RESTRICT:
			err = ctf_hash_insert(&fp->ctf_names, fp,
			    LCTF_INDEX_TO_TYPE(fp, id, child), tp->ctt_name);
			if (err != 0 && err != ECTF_STRTAB)
				return (err);
			break;
		}

		*xp = (uint32_t)((uintptr_t)tp - (uintptr_t)fp->ctf_buf);
		tp = (ctf_type_t *)((uintptr_t)tp + increment + vbytes);
	}

	ctf_dprintf("%lu total types processed\n", fp->ctf_typemax);
	ctf_dprintf("%u enum names hashed\n", ctf_hash_size(&fp->ctf_enums));
	ctf_dprintf("%u struct names hashed (%d long)\n",
	    ctf_hash_size(&fp->ctf_structs), nlstructs);
	ctf_dprintf("%u union names hashed (%d long)\n",
	    ctf_hash_size(&fp->ctf_unions), nlunions);
	ctf_dprintf("%u base type names hashed\n",
	    ctf_hash_size(&fp->ctf_names));

	/*
	 * Make an additional pass through the pointer table to find pointers
	 * that point to anonymous typedef nodes.  If we find one, modify the
	 * pointer table so that the pointer is also known to point to the
	 * node that is referenced by the anonymous typedef node.
	 */
	for (id = 1; id <= fp->ctf_typemax; id++) {
		if ((dst = fp->ctf_ptrtab[id]) != 0) {
			tp = LCTF_INDEX_TO_TYPEPTR(fp, id);

			if (LCTF_INFO_KIND(fp, tp->ctt_info) == CTF_K_TYPEDEF &&
			    strcmp(ctf_strptr(fp, tp->ctt_name), "") == 0 &&
			    LCTF_TYPE_ISCHILD(fp, tp->ctt_type) == child &&
			    LCTF_TYPE_TO_INDEX(fp, tp->ctt_type) <= fp->ctf_typemax)
				fp->ctf_ptrtab[
				    LCTF_TYPE_TO_INDEX(fp, tp->ctt_type)] = dst;
		}
	}

	return (0);
}

/*
 * Decode the specified CTF buffer and optional symbol table and create a new
 * CTF container representing the symbolic debugging information.  This code
 * can be used directly by the debugger, or it can be used as the engine for
 * ctf_fdopen() or ctf_open(), below.
 */
ctf_file_t *
ctf_bufopen(const ctf_sect_t *ctfsect, const ctf_sect_t *symsect,
    const ctf_sect_t *strsect, int *errp)
{
	const ctf_preamble_t *pp;
	ctf_header_t hp;
	ctf_file_t *fp;
	void *buf, *base;
	size_t size, hdrsz;
	int err;

	if (ctfsect == NULL || ((symsect == NULL) != (strsect == NULL)))
		return (ctf_set_open_errno(errp, EINVAL));

	if (symsect != NULL && symsect->cts_entsize != sizeof (Elf32_Sym) &&
	    symsect->cts_entsize != sizeof (Elf64_Sym))
		return (ctf_set_open_errno(errp, ECTF_SYMTAB));

	if (symsect != NULL && symsect->cts_data == NULL)
		return (ctf_set_open_errno(errp, ECTF_SYMBAD));

	if (strsect != NULL && strsect->cts_data == NULL)
		return (ctf_set_open_errno(errp, ECTF_STRBAD));

	if (ctfsect->cts_size < sizeof (ctf_preamble_t))
		return (ctf_set_open_errno(errp, ECTF_NOCTFBUF));

	pp = (const ctf_preamble_t *)ctfsect->cts_data;

	ctf_dprintf("ctf_bufopen: magic=0x%x version=%u\n",
	    pp->ctp_magic, pp->ctp_version);

	/*
	 * Validate each part of the CTF header (either V1 or V2).
	 * First, we validate the preamble (common to all versions).  At that
	 * point, we know specific header version, and can validate the
	 * version-specific parts including section offsets and alignments.
	 */
	if (pp->ctp_magic != CTF_MAGIC)
		return (ctf_set_open_errno(errp, ECTF_NOCTFBUF));

	if (_dt_unlikely_((pp->ctp_version < CTF_VERSION_1) ||
		(pp->ctp_version > CTF_VERSION_2)))
		return (ctf_set_open_errno(errp, ECTF_CTFVERS));

	if ((symsect != NULL) && (pp->ctp_version != CTF_VERSION_2)) {
		/*
		 * The symtab can contain function entries which contain
		 * embedded ctf info.  We do not support dynamically
		 * upgrading such entries (none should exist in any case, since
		 * dwarf2ctf does not create them).
		 */
		ctf_dprintf("ctf_bufopen: CTF version %d symsect not "
		    "supported\n", pp->ctp_version);
		return (ctf_set_open_errno(errp, ECTF_NOTSUP));
	}

	if (ctfsect->cts_size < sizeof (ctf_header_t))
		return (ctf_set_open_errno(errp, ECTF_NOCTFBUF));

	bcopy(ctfsect->cts_data, &hp, sizeof (hp));
	hdrsz = sizeof (ctf_header_t);

	size = hp.cth_stroff + hp.cth_strlen;

	ctf_dprintf("ctf_bufopen: uncompressed size=%lu\n", (ulong_t)size);

	if (hp.cth_lbloff > size || hp.cth_objtoff > size ||
	    hp.cth_funcoff > size || hp.cth_typeoff > size ||
	    hp.cth_stroff > size)
		return (ctf_set_open_errno(errp, ECTF_CORRUPT));

	if (hp.cth_lbloff > hp.cth_objtoff ||
	    hp.cth_objtoff > hp.cth_funcoff ||
	    hp.cth_funcoff > hp.cth_typeoff ||
	    hp.cth_funcoff > hp.cth_varoff ||
	    hp.cth_varoff > hp.cth_typeoff ||
	    hp.cth_typeoff > hp.cth_stroff)
		return (ctf_set_open_errno(errp, ECTF_CORRUPT));

	if ((hp.cth_lbloff & 3) || (hp.cth_objtoff & 1) ||
	    (hp.cth_funcoff & 1) || (hp.cth_varoff & 3) ||
	    (hp.cth_typeoff & 3))
		return (ctf_set_open_errno(errp, ECTF_CORRUPT));

	/*
	 * Once everything is determined to be valid, attempt to decompress
	 * the CTF data buffer if it is compressed.  Otherwise we just put
	 * the data section's buffer pointer into ctf_buf, below.
	 *
	 * Note: if this is a v1 buffer, it will be reallocated and expanded by
	 * init_types().
	 */
	if (hp.cth_flags & CTF_F_COMPRESS) {
		size_t srclen, dstlen;
		const void *src;
		int rc = Z_OK;

		if ((base = ctf_data_alloc(size + hdrsz)) == MAP_FAILED)
			return (ctf_set_open_errno(errp, ECTF_ZALLOC));

		bcopy(ctfsect->cts_data, base, hdrsz);
		((ctf_preamble_t *)base)->ctp_flags &= ~CTF_F_COMPRESS;
		buf = (uchar_t *)base + hdrsz;

		src = (uchar_t *)ctfsect->cts_data + hdrsz;
		srclen = ctfsect->cts_size - hdrsz;
		dstlen = size;

		if ((rc = uncompress(buf, &dstlen, src, srclen)) != Z_OK) {
			ctf_dprintf("zlib inflate err: %s\n", zError(rc));
			ctf_data_free(base, size + hdrsz);
			return (ctf_set_open_errno(errp, ECTF_DECOMPRESS));
		}

		if (dstlen != size) {
			ctf_dprintf("zlib inflate short -- got %lu of %lu "
			    "bytes\n", (ulong_t)dstlen, (ulong_t)size);
			ctf_data_free(base, size + hdrsz);
			return (ctf_set_open_errno(errp, ECTF_CORRUPT));
		}

	} else {
		base = (void *)ctfsect->cts_data;
		buf = (uchar_t *)base + hdrsz;
	}

	/*
	 * Once we have uncompressed and validated the CTF data buffer, we can
	 * proceed with allocating a ctf_file_t and initializing it.
	 *
	 * Nothing that depends on buf or base should be set directly in this
	 * function before the init_types() call, because it may be reallocated
	 * during transparent upgrade: see ctf_set_base() and
	 * ctf_realloc_base().
	 */
	if ((fp = ctf_alloc(sizeof (ctf_file_t))) == NULL)
		return (ctf_set_open_errno(errp, ENOMEM));

	bzero(fp, sizeof (ctf_file_t));
	ctf_set_version(fp, &hp, hp.cth_version);

	if (_dt_unlikely_(hp.cth_version < CTF_VERSION_2))
		fp->ctf_parmax = CTF_MAX_PTYPE_V1;
	else
		fp->ctf_parmax = CTF_MAX_PTYPE;

	bcopy(ctfsect, &fp->ctf_data, sizeof (ctf_sect_t));

	if (symsect != NULL) {
		bcopy(symsect, &fp->ctf_symtab, sizeof (ctf_sect_t));
		bcopy(strsect, &fp->ctf_strtab, sizeof (ctf_sect_t));
	}

	if (fp->ctf_data.cts_name != NULL)
		fp->ctf_data.cts_name = ctf_strdup(fp->ctf_data.cts_name);
	if (fp->ctf_symtab.cts_name != NULL)
		fp->ctf_symtab.cts_name = ctf_strdup(fp->ctf_symtab.cts_name);
	if (fp->ctf_strtab.cts_name != NULL)
		fp->ctf_strtab.cts_name = ctf_strdup(fp->ctf_strtab.cts_name);

	if (fp->ctf_data.cts_name == NULL)
		fp->ctf_data.cts_name = _CTF_NULLSTR;
	if (fp->ctf_symtab.cts_name == NULL)
		fp->ctf_symtab.cts_name = _CTF_NULLSTR;
	if (fp->ctf_strtab.cts_name == NULL)
		fp->ctf_strtab.cts_name = _CTF_NULLSTR;

	if (strsect != NULL) {
		fp->ctf_str[CTF_STRTAB_1].cts_strs = strsect->cts_data;
		fp->ctf_str[CTF_STRTAB_1].cts_len = strsect->cts_size;
	}

	ctf_set_base(fp, &hp, base);
	fp->ctf_size = size + hdrsz;

	if ((err = init_types(fp, &hp)) != 0) {
		(void) ctf_set_open_errno(errp, err);
		goto bad;
	}

	/*
	 * The ctf region may have been reallocated by init_types(), but now
	 * that is done, it will not move again, so we can protect it, as long
	 * as it didn't come from the ctfsect, wihcih might have been allocated
	 * with malloc().
	 */
	if (fp->ctf_base != (void *)ctfsect->cts_data)
		ctf_data_protect((void *) fp->ctf_base, fp->ctf_size);

	/*
	 * If we have a symbol table section, allocate and initialize
	 * the symtab translation table, pointed to by ctf_sxlate.
	 */
	if (symsect != NULL) {
		fp->ctf_nsyms = symsect->cts_size / symsect->cts_entsize;
		fp->ctf_sxlate = ctf_alloc(fp->ctf_nsyms * sizeof (uint32_t));

		if (fp->ctf_sxlate == NULL) {
			(void) ctf_set_open_errno(errp, ENOMEM);
			goto bad;
		}

		if ((err = init_symtab(fp, &hp, symsect, strsect)) != 0) {
			(void) ctf_set_open_errno(errp, err);
			goto bad;
		}
	}

	/*
	 * Initialize the ctf_lookup_by_name top-level dictionary.  We keep an
	 * array of type name prefixes and the corresponding ctf_hash to use.
	 * NOTE: This code must be kept in sync with the code in ctf_update().
	 */
	fp->ctf_lookups[0].ctl_prefix = "struct";
	fp->ctf_lookups[0].ctl_len = strlen(fp->ctf_lookups[0].ctl_prefix);
	fp->ctf_lookups[0].ctl_hash = &fp->ctf_structs;
	fp->ctf_lookups[1].ctl_prefix = "union";
	fp->ctf_lookups[1].ctl_len = strlen(fp->ctf_lookups[1].ctl_prefix);
	fp->ctf_lookups[1].ctl_hash = &fp->ctf_unions;
	fp->ctf_lookups[2].ctl_prefix = "enum";
	fp->ctf_lookups[2].ctl_len = strlen(fp->ctf_lookups[2].ctl_prefix);
	fp->ctf_lookups[2].ctl_hash = &fp->ctf_enums;
	fp->ctf_lookups[3].ctl_prefix = _CTF_NULLSTR;
	fp->ctf_lookups[3].ctl_len = strlen(fp->ctf_lookups[3].ctl_prefix);
	fp->ctf_lookups[3].ctl_hash = &fp->ctf_names;
	fp->ctf_lookups[4].ctl_prefix = NULL;
	fp->ctf_lookups[4].ctl_len = 0;
	fp->ctf_lookups[4].ctl_hash = NULL;

	if (symsect != NULL) {
		if (symsect->cts_entsize == sizeof (Elf64_Sym))
			(void) ctf_setmodel(fp, CTF_MODEL_LP64);
		else
			(void) ctf_setmodel(fp, CTF_MODEL_ILP32);
	} else
		(void) ctf_setmodel(fp, CTF_MODEL_NATIVE);

	fp->ctf_refcnt = 1;
	return (fp);

bad:
	ctf_close(fp);
	return (NULL);
}

/*
 * Close the specified CTF container and free associated data structures.  Note
 * that ctf_close() is a reference counted operation: if the specified file is
 * the parent of other active containers, its reference count will be greater
 * than one and it will be freed later when no active children exist.
 */
void
ctf_close(ctf_file_t *fp)
{
	ctf_dtdef_t *dtd, *ntd;
	ctf_dvdef_t *dvd, *nvd;

	if (fp == NULL)
		return; /* allow ctf_close(NULL) to simplify caller code */

	ctf_dprintf("ctf_close(%p) refcnt=%u\n", (void *)fp, fp->ctf_refcnt);

	if (fp->ctf_refcnt > 1) {
		fp->ctf_refcnt--;
		return;
	}

	if(fp->ctf_dynparname != NULL)
		ctf_free(fp->ctf_dynparname, strlen(fp->ctf_dynparname) + 1);

	if (fp->ctf_parent != NULL)
		ctf_close(fp->ctf_parent);

	for (dtd = ctf_list_next(&fp->ctf_dtdefs); dtd != NULL; dtd = ntd) {
		ntd = ctf_list_next(dtd);
		ctf_dtd_delete(fp, dtd);
	}

	ctf_free(fp->ctf_dthash, fp->ctf_dthashlen * sizeof (ctf_dtdef_t *));

	for (dvd = ctf_list_next(&fp->ctf_dvdefs); dvd != NULL; dvd = nvd) {
		nvd = ctf_list_next(dvd);
		ctf_dvd_delete(fp, dvd);
	}

	ctf_free(fp->ctf_dvhash, fp->ctf_dvhashlen * sizeof (ctf_dvdef_t *));

	if (fp->ctf_flags & LCTF_MMAP) {
		if (fp->ctf_data.cts_data != NULL)
			ctf_sect_munmap(&fp->ctf_data);
		if (fp->ctf_symtab.cts_data != NULL)
			ctf_sect_munmap(&fp->ctf_symtab);
		if (fp->ctf_strtab.cts_data != NULL)
			ctf_sect_munmap(&fp->ctf_strtab);
	}

	if (fp->ctf_data.cts_name != _CTF_NULLSTR &&
	    fp->ctf_data.cts_name != NULL) {
		ctf_free((char *)fp->ctf_data.cts_name,
		    strlen(fp->ctf_data.cts_name) + 1);
	}

	if (fp->ctf_symtab.cts_name != _CTF_NULLSTR &&
	    fp->ctf_symtab.cts_name != NULL) {
		ctf_free((char *)fp->ctf_symtab.cts_name,
		    strlen(fp->ctf_symtab.cts_name) + 1);
	}

	if (fp->ctf_strtab.cts_name != _CTF_NULLSTR &&
	    fp->ctf_strtab.cts_name != NULL) {
		ctf_free((char *)fp->ctf_strtab.cts_name,
		    strlen(fp->ctf_strtab.cts_name) + 1);
	}

	ctf_free_base(fp, NULL, 0);

	if (fp->ctf_sxlate != NULL)
		ctf_free(fp->ctf_sxlate, sizeof (uint32_t) * fp->ctf_nsyms);

	if (fp->ctf_txlate != NULL) {
		ctf_free(fp->ctf_txlate,
		    sizeof (uint32_t) * (fp->ctf_typemax + 1));
	}

	if (fp->ctf_ptrtab != NULL) {
		ctf_free(fp->ctf_ptrtab,
		    sizeof (uint32_t) * (fp->ctf_typemax + 1));
	}

	ctf_hash_destroy(&fp->ctf_structs);
	ctf_hash_destroy(&fp->ctf_unions);
	ctf_hash_destroy(&fp->ctf_enums);
	ctf_hash_destroy(&fp->ctf_names);

	ctf_free(fp, sizeof (ctf_file_t));
}

/*
 * Return the ctfsect out of the core ctf_impl.  Useful for freeing the
 * ctfsect's data * after ctf_close(), which is why we return the actual
 * structure, not a pointer to it, since that is likely to become a pointer to
 * freed data before the return value is used under the expected use case of
 * ctf_getsect()/ ctf_close()/free().
 */
extern ctf_sect_t ctf_getdatasect(const ctf_file_t *fp)
{
	return (fp->ctf_data);
}

/*
 * Return the CTF handle for the parent CTF container, if one exists.
 * Otherwise return NULL to indicate this container has no imported parent.
 */
ctf_file_t *
ctf_parent_file(ctf_file_t *fp)
{
	return (fp->ctf_parent);
}

/*
 * Return the name of the parent CTF container, if one exists.  Otherwise
 * return NULL to indicate this container is a root container.
 */
const char *
ctf_parent_name(ctf_file_t *fp)
{
	return (fp->ctf_parname);
}

/*
 * Set the parent name.  It is an error to call this routine without calling
 * ctf_import() at some point.
 */
void
ctf_parent_name_set(ctf_file_t *fp, const char *name)
{
	if(fp->ctf_dynparname != NULL)
		ctf_free(fp->ctf_dynparname, strlen(fp->ctf_dynparname) + 1);

	fp->ctf_dynparname = ctf_strdup(name);
	fp->ctf_parname = fp->ctf_dynparname;
}

/*
 * Import the types from the specified parent container by storing a pointer
 * to it in ctf_parent and incrementing its reference count.  Only one parent
 * is allowed: if a parent already exists, it is replaced by the new parent.
 */
int
ctf_import(ctf_file_t *fp, ctf_file_t *pfp)
{
	if (fp == NULL || fp == pfp || (pfp != NULL && pfp->ctf_refcnt == 0))
		return (ctf_set_errno(fp, EINVAL));

	if (pfp != NULL && pfp->ctf_dmodel != fp->ctf_dmodel)
		return (ctf_set_errno(fp, ECTF_DMODEL));

	if (fp->ctf_parent != NULL)
		ctf_close(fp->ctf_parent);

	if (pfp != NULL) {
		fp->ctf_flags |= LCTF_CHILD;
		pfp->ctf_refcnt++;

		if (fp->ctf_parname == NULL)
			ctf_parent_name_set(fp, "PARENT");
	}
	fp->ctf_parent = pfp;
	return (0);
}

/*
 * Set the data model constant for the CTF container.
 */
int
ctf_setmodel(ctf_file_t *fp, int model)
{
	const ctf_dmodel_t *dp;

	for (dp = _libctf_models; dp->ctd_name != NULL; dp++) {
		if (dp->ctd_code == model) {
			fp->ctf_dmodel = dp;
			return (0);
		}
	}

	return (ctf_set_errno(fp, EINVAL));
}

/*
 * Return the data model constant for the CTF container.
 */
int
ctf_getmodel(ctf_file_t *fp)
{
	return (fp->ctf_dmodel->ctd_code);
}

void
ctf_setspecific(ctf_file_t *fp, void *data)
{
	fp->ctf_specific = data;
}

void *
ctf_getspecific(ctf_file_t *fp)
{
	return (fp->ctf_specific);
}
